import functools

from flask import current_app

__all__ = ["endpoint"]


def endpoint(fn):
    """
    Decorate a function to pass 'app' to it
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        app = kwargs.pop("app", current_app.flasket)
        return fn(
            app=app,
            *args,
            **kwargs,
        )

    return wrapper



"""
Classes
-------

HTTPException
^^^^^^^^^^^^^

.. autoclass:: HTTPException
  :show-inheritance:
  :members:

.. HTTPExceptions
.. ^^^^^^^^^^^^^^
..
.. .. autoclass:: HTTPExceptions
..   :noindex:
..   :members:

NoContent
^^^^^^^^^

.. autoclass:: NoContent
  :show-inheritance:
  :members:

BadRequest
^^^^^^^^^^

.. autoclass:: BadRequest
  :show-inheritance:
  :members:

  *(Content generated from Flask documentation)*

Unauthorized
^^^^^^^^^^^^

.. autoclass:: Unauthorized
  :show-inheritance:
  :members:

  *(Content generated from Flask documentation)*

Forbidden
^^^^^^^^^

.. autoclass:: Forbidden
  :show-inheritance:
  :members:

  *(Content generated from Flask documentation)*

NotFound
^^^^^^^^

.. autoclass:: NotFound
  :show-inheritance:
  :members:

  *(Content generated from Flask documentation)*

FailedDependency
^^^^^^^^^^^^^^^^

.. autoclass:: FailedDependency
  :show-inheritance:
  :members:

  *(Content generated from Flask documentation)*

InternalServerError
^^^^^^^^^^^^^^^^^^^

.. autoclass:: InternalServerError
  :show-inheritance:
  :members:

  *(Content generated from Flask documentation)*

NotImplemented
^^^^^^^^^^^^^^

.. autoclass:: NotImplemented
  :show-inheritance:
  :members:

  *(Content generated from Flask documentation)*
"""

from werkzeug import exceptions as exception

__all__ = [
    "BadRequest",
    "FailedDependency",
    "Forbidden",
    "HTTPException",
    "InternalServerError",
    "NoContent",
    "NotFound",
    "NotImplemented",
    "ServiceUnavailable",
    "Unauthorized",
]

# Notes on exceptions handling and errors:
# https://werkzeug.palletsprojects.com/en/master/exceptions/
# https://tools.ietf.org/html/rfc7807
HTTPException = exception.HTTPException


# 204 No Content
class NoContent(HTTPException):
    """
    *204 No Content*

    Indicates that the server has successfully fulfilled the request and that there is no content to send in the response payload body.
    """

    code = 204
    name = "No Content"


# 400 Bad Request
BadRequest = exception.BadRequest

# 401 Unauthorized
Unauthorized = exception.Unauthorized

# 403 Forbidden
Forbidden = exception.Forbidden

# 404 NotFound
NotFound = exception.NotFound

# 424 FailedDependency
FailedDependency = exception.FailedDependency

# 500 InternalServerError
InternalServerError = exception.InternalServerError

# 501 NotImplemented
# pylint: disable=redefined-builtin
NotImplemented = exception.NotImplemented

# 503 ServiceUnavailable
ServiceUnavailable = exception.ServiceUnavailable


class HTTPExceptions:
    """
    Minimal class for inheritance by :ref:`flasket.properties.FlasketProperties` class

    This class recieves all members of the ``flasket.exceptions`` module.
    """


# Copy all symbols above into HTTPExceptions
for x in __all__:
    setattr(HTTPExceptions, x, locals()[x])



"""
Description
-----------

Main application. It handles returning the WSGI application,
but also returning the potential services it can connect to and use.

On the OpenAPI side, it handles merging multiple yaml files into a single
specification before loading it.

Classes
-------

Flasket
^^^^^^^

.. autoclass:: Flasket
  :show-inheritance:
  :members: baseurl, clients, config, current_app, debug, flask, backends, host,  make_response, port, redirect, render_template, request, rootpath, session, sitename,

"""
import logging
import os
import random
import string
import sys
import typing as t
import weakref

from flask import Flask, current_app
from flask import g as flask_g
from flask import request as current_request
from flask_gordon.ext.defaults import SERVER_LISTEN, SERVER_PORT
from flask_gordon.ext.defaults import make_config as make_settings
from flask_gordon.middleware import BeforeFirstCall, Dispatcher
from werkzeug.middleware.proxy_fix import ProxyFix as ProxyMiddleware

from flasket.middleware import GunicornApplication

from .backends import ApiBackend, DashBackend, StaticBackend
from .logger import Logger
from .properties import LoggerProperties

__all__: t.List[str] = [
    "Flasket",
]


class Flasket(
    Flask,
    LoggerProperties,
):
    def __init__(self, name):
        self._flasket = weakref.proxy(self)
        super().__init__(name)

        Logger.configure()
        self.logger.setLevel(logging.INFO)

        self._middleware = None
        self._wsgi_app = self.wsgi_app
        self._rootpath = os.path.dirname(__file__)

        # Backends loaded, and to load
        self._backends = {}
        self._backends_cls = [DashBackend, ApiBackend, StaticBackend]

        # Did we go through self.configure
        self._configured = False
        self.config["SETTINGS"] = make_settings()

    def _init_debug(self) -> None:
        """
        Set the class variables to the values that were passed on by configuration.
        If debug is on, set the logger level
        """
        debug = self.settings["server"].get("debug", False)
        if debug:
            self.logger.setLevel(logging.DEBUG)

    def _init_rootpath(self, rootpath: str) -> None:
        def add_to_syspath(path):
            if path not in sys.path:
                self.logger.info(f'Adding path "{path}" to sys.path to enable dynamic loading...')
                sys.path.insert(0, path)

        if not rootpath:
            rootpath = os.getcwd()

        # rootpath was probably passed by "__file__"
        if rootpath and os.path.isfile(rootpath):
            rootpath = os.path.dirname(rootpath)
        rootpath = os.path.abspath(rootpath)

        if not os.path.exists(rootpath) or not os.path.isdir(rootpath):
            self.logger.error(f'Root path directory "{rootpath}" does not exist.')
            sys.exit(1)
        self.logger.info(f'Using root path "{rootpath}" for dynamic loading.')

        add_to_syspath(rootpath)
        self._rootpath = rootpath

    def _init_middleware(self, middleware: str = None) -> None:
        if middleware == "flask":
            return
        if middleware == "gunicorn":
            self._middleware = GunicornApplication(self, self.config["SETTINGS"])
            return
        raise ValueError("unknown middleware")

    def _init_defaults(self) -> None:
        # Configuration comes from several locations:
        # - the command line arguments,
        # - the configuration file
        # - defaults
        #
        # Flask is also very flexible to where settings can come from.
        #
        # The most important variables, mostly to start/pre-configure the service
        # have already been set
        #
        # Set some new defaults
        self.config["JSON_SORT_KEYS"] = False
        self.config["TEMPLATES_AUTO_RELOAD"] = not self.settings["server"].get("production")
        self.config["EXPLAIN_TEMPLATE_LOADING"] = os.getenv("EXPLAIN_TEMPLATE_LOADING")
        self.config["UNIQUE_KEY"] = "".join(random.choice(string.ascii_lowercase) for i in range(6))

    # FIXME: flask / server?
    def _init_config(self) -> None:
        server = self.settings.get("flask") or {}
        for key, _ in self.config.items():
            lkey = key.lower()
            if lkey in server:
                self.config[key] = server[lkey]

    # FIXME: Move to flask-gordon
    def _init_secret(self) -> None:
        # Handle the secret session key if missing
        secret_key = self.config["SECRET_KEY"]
        if not secret_key:
            self.config["SECRET_KEY"] = "".join(random.choices(string.ascii_letters, k=20))
            self.logger.warning("Generated a random secret session key.")

    def _init_backends(self, options) -> None:
        # Load all backends
        # For every backend get option[name] for that backend
        options = options or {}
        for cls in self._backends_cls:  # pylint: disable=not-an-iterable
            cls_options = options.get(cls.name()) or {}
            self.logger.info(f"Initializing {cls.name()} backend...")
            backend = cls.init_app(flasket=weakref.proxy(self), rootpath=self._rootpath, options=cls_options)
            self._backends[backend.prefix] = backend

    def _init_wsgi_apps(self) -> None:
        # Queue the middlewares
        middleware = self.wsgi_app
        middleware = Dispatcher(self._backends["/"], self._backends)
        proxy = self.settings["server"].get("proxy", {})
        if proxy:
            middleware = ProxyMiddleware(middleware, **proxy)
        middleware = BeforeFirstCall(middleware, self._before_first_call)
        self.wsgi_app = middleware

    def configure(
        self,
        rootpath: str,
        middleware: str = None,
        options=None,
    ):
        assert self._configured is False, "You can only run 'Flasket.configure' once"
        self._configured = True

        self.logger.info("Configuring Flasket application...")

        self._init_debug()
        self._init_rootpath(rootpath)
        self._init_middleware(middleware)
        self._init_defaults()
        self._init_config()
        self._init_secret()

        # https://flask.palletsprojects.com/en/2.2.x/config/
        # Values from _init_defaults could have been overloaded
        # by _init_config
        self.config["DEBUG"] = self.settings["server"].get("debug", False)

        self._init_backends(options)
        self._init_wsgi_apps()
        return self

    # pylint: disable=arguments-differ
    def run(self):
        if not self._configured:
            self.configure(__file__)

        if self._middleware is None:
            host = self.config["SETTINGS"]["server"].get("listen", SERVER_LISTEN)
            port = self.config["SETTINGS"]["server"].get("port", SERVER_PORT)
            debug = self.config["SETTINGS"]["server"].get("debug", False)
            super().run(host=host, port=port, debug=debug, load_dotenv=False, use_reloader=False)
        else:
            self._middleware.run()

    def _register_backend(self, backend):
        # Copy all flask config to backend config
        for key, value in self.config.items():
            # blacklist some keys
            try:
                # pylint: disable=protected-access
                backend._flask.config[key] = value
            except AttributeError:
                continue

        # And add 'flasket' to Flask app, so it's available via current_app
        # pylint: disable=protected-access
        backend._flask.flasket = weakref.proxy(self)
        backend._flask.before_request(Flasket._before_request)
        backend._flask.after_request(Flasket._after_request)

    # --------------------------------------------------------------------------
    @staticmethod
    def _before_first_call() -> None:
        Logger.disable()

    @staticmethod
    def _before_request() -> None:
        # Inject some variables available to templates
        flasket = current_app.flasket
        flask_g.flasket = flasket

    @staticmethod
    def _after_request(response) -> None:
        # Log the request
        Logger.log_request(current_request, response)
        return response

    # --------------------------------------------------------------------------
    def route(self, *args, **kwargs):
        raise NotImplementedError('to "flasket.route" is not implemented. Use "add_url_rule" instead.')

    # --------------------------------------------------------------------------
    @property
    def user_options(self, *args) -> None:
        """
        Network post on which Flasket is listening on.
        """
        raise NotImplementedError('Incorrect application, try appending ".celery"')

    @property
    def logger(self) -> logging.Logger:
        """
        Return a :class:`logging.Logger` named ``stderr``
        """
        return logging.getLogger("stderr")

    @property
    def settings(self) -> t.Dict[str, t.Any]:
        """
        Return Flasket settings used at start
        """
        return self.config.get("SETTINGS") or {}

    @property
    def rootpath(self) -> str:
        """
        Return rootpath
        """
        return self._rootpath

    @property
    def host(self) -> str:
        """
        Network hosts on which Flasket is listening on.
        """
        return self.settings["server"].get("listen", SERVER_LISTEN)

    @property
    def port(self) -> int:
        """
        Network post on which Flasket is listening on.
        """
        return self.settings["server"].get("port", 8080)

    @property
    def sitename(self) -> str:
        """
        Return combination of network host and port
        """
        if self.port == 80:
            return f"http://{self.host}"
        return f"http://{self.host}:{self.port}"

    @property
    def baseurl(self) -> str:
        """
        Returns baseurl "{scheme}://{netloc}"
        """
        baseurl = self.settings["server"].get("baseurl", None)
        if baseurl:
            return baseurl
        return self.sitename

    @property
    def backends(self) -> t.Dict[str, "Backend"]:
        """
        Returns array of backends
        """
        return self._backends

    @property
    def production(self) -> bool:
        """
        Returns production status. Not equivalent to `not debug`
        """
        return self.settings["server"].get("production", False)



import collections
import os
import sys
import traceback
import typing as t
from copy import deepcopy

import yaml
from boltons.fileutils import iter_find_files
from connexion import FlaskApp
from connexion.exceptions import InvalidSpecification

from .._utils import deepmerge
from ._backend import Backend


def _load_yamlfiles(files: t.List[str]) -> t.Dict:
    """
    For every yaml file, load the file and merge with previous

    Parameters
    ----------
    files: List[str]
        list of files to parse and merge

    Returns
    -------
    dict:
        merged dictionary
    """
    try:
        data = {}
        for file in files:
            with open(file, encoding="UTF-8") as fd:
                newdata = yaml.safe_load(fd)
            if newdata:
                data = deepmerge(data, newdata)
        return data
    except yaml.scanner.ScannerError as err:
        raise type(err)(
            f"yaml scanner error: {err.problem} in '{err.problem_mark.name}', at line {err.problem_mark.line}",
        ) from None
    except yaml.parser.ParserError as err:
        raise type(err)(
            f"yaml parser error: {err.problem} in '{err.problem_mark.name}', at line {err.problem_mark.line}",
        ) from None
    except Exception as err:
        raise type(err)(f"{str(err)} in file {file}") from None


def _find_yamlfiles(paths: t.List[str], logger: object) -> collections.OrderedDict:
    """
    Find all yaml files in paths, returns a merged dict

    Parameters
    ----------
    paths: List[str]
        list of paths to search

    Returns
    -------
    dict:
        merged dictionary
    """
    # Make paths unique but keep order
    paths = list(dict.fromkeys(paths))

    # Merge all yml files from the directory
    files = []
    for path in paths:
        if os.path.exists(path):
            if logger:
                logger.info(f'Reading OpenAPI files from "{path}"...')
            files.extend(iter_find_files(path, ["*.yml", "*.yaml"]))

    # Load the contents
    data = _load_yamlfiles(files)

    # sort the keys in /components/schemas, since the UI uses that order
    data["components"]["schemas"] = collections.OrderedDict(
        sorted((data["components"]["schemas"]).items(), key=lambda s: s[0].lower()),
    )
    data["paths"] = collections.OrderedDict(
        sorted((data["paths"]).items(), key=lambda s: s[0].lower()),
    )
    return data


def _remove_path_if_key(specs: t.Dict, key: str, value: str) -> t.Dict:
    """
    For every API path in specs, remove it if we find specs[...][key] = value

    Parameters
    ----------
    specs: Dict
        OpenAPI specs

    key: str
        Key to look for

    value: str
        Value to look for

    Returns
    -------
    dict:
        OpenAPI specs without the removed keys
    """
    rv = deepcopy(specs)
    for path in specs.get("paths", {}).keys():
        for method in specs["paths"][path].keys():
            if key not in specs["paths"][path][method]:
                continue
            if specs["paths"][path][method][key] == value:
                del rv["paths"][path][method]
    return rv


class ApiBackend(Backend):
    @staticmethod
    def init_app(flasket, rootpath, options=None) -> "ApiBackend":
        return ApiBackend(flasket, rootpath, options)

    def __init__(self, flasket, rootpath, _options=None):
        super().__init__(flasket, rootpath)

        self._specs = None

        self._init_rootpath()
        self._init_specs()
        self._init_connexion()
        self._flasket._register_backend(self)

    def _init_rootpath(self) -> None:
        self._rootpath = os.path.join(self._rootpath, "api")
        if not os.path.exists(self._rootpath):
            self.logger.warning('Directory "api" in root path does not exist.')

    def _init_specs(self) -> None:
        # Build specs from yml files, and remove debug endpoints
        specpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "../api"))
        searchpath = [specpath, self._rootpath]
        specs = _find_yamlfiles(searchpath, self.logger)
        # Don't get flask.debug, but settings debug
        if not self.flasket.settings["server"].get("debug", False):
            self.logger.info("Removing API debug endpoints.")
            specs = _remove_path_if_key(specs, "x-debug", True)
        else:
            self.logger.warning("Keeping API debug endpoints.")
        self._specs = specs

    def _init_connexion(self) -> None:
        # https://connexion.readthedocs.io/en/latest/
        self._connexion = FlaskApp(__name__, specification_dir="api")
        enable_ui = (self.flasket.settings["server"].get("ui", True),)
        try:
            self._connexion.add_api(
                self._specs,
                strict_validation=True,
                validate_responses=True,
                options={
                    "openapi_spec_path": "/api/openapi.json",
                    "swagger_url": "/api/ui",
                    "swagger_ui": enable_ui,
                    "swagger_spec": enable_ui,
                },
            )
            self._flask = self._connexion.app

        except InvalidSpecification as err:
            raise type(err)(f"OpenAPI specification error: {err.message}") from None
        except Exception:
            # TODO: improve on this (raise as import error for example)
            print(traceback.print_exc(), file=sys.stderr)
            raise

        if enable_ui:
            self.logger.info(f"Swagger UI is available at: {self.flasket.sitename}/api/ui")
        else:
            self.logger.info("Swagger UI not started")

    @staticmethod
    def name() -> str:
        return "api"

    @property
    def prefix(self) -> str:
        return "/api"
